
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Manipulate iTunes/iPod playlists
'' Stephen Clarke-Willson, Ph.D.
'' First created 2006 11 28
'' Last mod 2008 01 15
'' Released to the public domain
''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Apple released some sample code in Javascript but (duh) I didn't know
'' you could just run the Javascript in Windows, so I started making
'' VBS scripts that accessed the iTunes COM library, and that's how this
'' got started.

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Ensure we are running CScript
'' If you get this error, then run
'' wscript //H:CScript
'' from a command line to convert to using CScript as your
'' default; or just run this script with
'' CScript Playlists.vbs <arguments>

If UCase(Right(Wscript.FullName, 11)) = "WSCRIPT.EXE" Then
    Wscript.Echo "This script must be run under CScript.  Use 'cscript playlists.vbs <arguments>' at a command prompt."
    Wscript.Quit
End If


'' Tell scripting engine to ignore errors so we can handle them
'' directly; you might want to comment this out when debugging.

on error Resume Next

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Define some constants (which Javascript would just know, I think)

rem enum  ITSourceKind { 
    const ITSourceKindUnknown =         0
    const ITSourceKindLibrary =         1
    const ITSourceKindIPod =            2
    const ITSourceKindAudioCD =         3 
    const ITSourceKindMP3CD =           4
    const ITSourceKindDevice =          5
    const ITSourceKindRadioTuner =      6    
    const ITSourceKindSharedLibrary =   7
rem } 

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
rem enum  ITPlaylistKind { 
    const ITPlaylistKindUnknown =       0
    const ITPlaylistKindLibrary =       1 
    const ITPlaylistKindUser =          2 
    const ITPlaylistKindCD =            3
    const ITPlaylistKindDevice =        4
    const ITPlaylistKindRadioTuner =    5 
rem } 

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Define our options

'' Select one of iPod or iTunes
optionSourceType                = ITSourceKindUnknown

'' Select whether to include smart playlists when scanning for
'' unreferenced tracks
optionIgnoreSmartPlaylists      = false

'' Select whether to delete unreferenced tracks
optionDeleteUnreferenced        = false

'' Select whether to move unreferenced tracks to "Unreferenced" 
'' playlist.
optionMoveToUnreferencedList    = false

'' Select whether to spew extra status information
optionVerbose                   = false

'' Select whether to force delete all tracks on the 
'' playlist "DeleteMe".  You would drag tracks you want
'' deleted to this playlist and then force-delete them.
optionDeleteDeleteMeList        = false

'' Select whether to export a list of all tracks
'' and their ratings to StdOut. Use redirection to
'' write a file (e.g., PlayLists.vbs iTunes exportRatings >ratings.txt)
optionExportRatings             = false

'' Select whether to import a list of tracks and their
'' ratings.  Reads from StdIn.
optionImportRatings             = false

'' Select whether to match the name of a track when importing
'' a rating.
optionCheckName                 = true

'' Select whether to match the album name of a track when
'' importing a rating.
optionCheckAlbum                = true

'' Select whether to match the artist name of a track when
'' importing a rating.
optionCheckArtist               = true

'' Select whether to match the size of a track when importing
'' a rating.
optionCheckSize                 = true

'' Select whether to copy the album name as well as the rating
'' when importing ratings.  It doesn't make too much sense to
'' do this if also checking against the album name.
optionCopyAlbum                 = false

'' Select whether to copy the artist name as well as the rating
'' when importing ratings.  It doesn't make too much sense to
'' do this if also checking against the artist name.
optionCopyArtist                = false

'' Select whether to just print that the code will do something
'' as opposed to actually doing it, for testing.  Not consistently
'' implemented, I think.
optionTestOnly                  = false


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Parse our options
Set objArgs = WScript.Arguments
For Each strArg in objArgs
    if StrComp(strArg, "iPod", 1) = 0 then
        optionSourceType = ITSourceKindIPod
    elseif StrComp(strArg, "iTunes", 1) = 0 then
        optionSourceType = ITSourceKindLibrary
    elseif StrComp(strArg, "moveToUnref", 1) = 0 then
        optionMoveToUnreferencedList = true
    elseif StrComp(strArg, "delUnreferenced", 1) = 0 then
        optionDeleteUnreferenced = true
    elseif StrComp(strArg, "deleteDeleteMeList", 1) = 0 then
        optionDeleteDeleteMeList = true
    elseif StrComp(strArg, "verbose", 1) = 0 then
        optionVerbose = true
    elseif StrComp(strArg, "ignoreSmartPlaylists", 1) = 0 then
        optionIgnoreSmartPlaylists = true
    elseif StrComp(strArg, "exportRatings", 1) = 0 then
        optionExportRatings = true
    elseif StrComp(strArg, "importRatings", 1) = 0 then
        optionImportRatings = true
    elseif StrComp(strArg, "skipName", 1) = 0 then
        optionCheckName = false
    elseif StrComp(strArg, "skipAlbum", 1) = 0 then
        optionCheckAlbum = false
    elseif StrComp(strArg, "skipArtist", 1) = 0 then
        optionCheckArtist = false
    elseif StrComp(strArg, "skipSize", 1) = 0 then
        optionCheckSize = false
    elseif StrComp(strArg, "copyAlbum", 1) = 0 then
        optionCopyAlbum = true
    elseif StrComp(strArg, "copyArtist", 1) = 0 then
        optionCopyArtist = true
    elseif StrComp(strArg, "testOnly", 1) = 0 then
        optionTestOnly = true
    else
        WScript.Echo "Unknown command argument: " & strArg
    end if
Next 

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Verify iTunes or iPod specified.  I don't support multiple iPods.
if optionSourceType = ITSourceKindUnknown then
    WScript.Echo "You must specify iPod or iTunes"
    WScript.Quit
end if

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' I couldn't store an object reference in a dictionary entry, so I store the index
'' of the object in the trackObjects array instead.
Dim trackObjects(100000)
objectIdx = 0

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
set iTunesApp               = WScript.CreateObject("iTunes.Application")
'set    mainLibrary         = iTunesApp.LibraryPlaylist
'set    mainLibrarySource   = iTunesApp.LibrarySource
'set    mainLibraryTracks   = mainLibrary.Tracks

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
set sources = iTunesApp.Sources
sourceCount = sources.Count

'' Grab first source object of the type specified on the command line.
for i = 1 to sourceCount
    kind = sources.Item(i).Kind
    if kind = optionSourceType then
        set src = sources.Item(i)
    end if
next

if not IsObject(src) then
    WScript.Echo "Couldn't find iPod or iTunes"
    WScript.quit
end if

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
WScript.Echo "Working..."

set trackCollection = CreateObject("Scripting.Dictionary")
set playlists   = src.Playlists
playlistsCount  = playlists.Count


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Handle deleteMe option
'' Before I knew you could Shift-Delete things in iTunes, I had a helluva time actually
'' deleting stuff from my iPod.  I could only delete the item from a playlist, or go
'' find it in the main playlist and delete it.  So, I would drag things to a "DeleteMe"
'' playlist, and then run this script to remove things I no longer wanted on the iPod.
if optionDeleteDeleteMeList then
    for i = 2 to playlistsCount  ' the first playlist is everything on the iPod
        '' Brute force find DeleteMe playlist
        set playlist = playlists.Item(i)
        if playlist.Name = "DeleteMe" then
            WScript.Echo "Permanently deleting tracks in playlist 'DeleteMe'; any other options will be ignored."
            set tracks = playlist.Tracks
            count = tracks.Count
            do while tracks.Count <> 0      '' the list size can shrink by more than one each time through if there are dups in the list
                set track = tracks.Item(1)  '' the list keeps shrinking, so just keep deleting the first item
                trackDBId = track.TrackDatabaseID
                WScript.Echo "Deleting '" & track.Name & "'; Rating: " & track.Rating
                'WScript.Echo "track.TrackDatabaseID = " & trackDBId
                '''track.Delete  '' will only delete from this playlist since this is a user playlist
                '' So, since Delete method won't work, brute force search main library for this track
                set mainPlaylist = playlists.Item(1)
                mainPlaylistCount = mainPlaylist.Tracks.Count
                for q = 1 to mainPlaylistCount
                    set mainLibTrack = mainPlaylist.Tracks.Item(q)
                    'WSCript.Echo "Searching " & mainLibTrack.TrackDatabaseID & " '" & mainLibTrack.Name & "'"
                    if mainLibTrack.TrackDatabaseID = trackDBId then
                        'WSCript.Echo "Found dbID = " & trackDBId & "; name = '" & mainLibTrack.Name & "'; Rating: " & mainLibTrack.Rating
                        if not optionTestOnly then
                            mainLibTrack.Delete '' to truly delete something, you have to delete it from the main library track list
                        end if
                        exit for
                    end if
                next
            loop
            WScript.quit
        end if
    next
    WScript.Echo "DeleteMe playlist not found."
    WScript.quit
end if

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Ratings are stored in the iTunes library and not in the files, so if you
'' want to move ratings from one library to another, these next two routines
'' will get the job done for you.
''
'' Writes to StdOut so to save your ratings, do something like this:
'''''   Playlists.vbs iTunes exportRatings >ratings.txt
''
if optionExportRatings then
    '' first playlist is everything in the library
    set playlist = playlists.Item(1)
    set tracks = playlist.Tracks
    count = tracks.Count
    for i = 1 to count 
        set track = tracks.Item(i)
        '' Emit enough stuff to hopefully uniquely identify the track in another
        '' library
        WScript.Echo "*New*"
        WScript.Echo "name"
        WScript.Echo track.Name
        WScript.Echo "album"
        WScript.Echo track.Album
        WScript.Echo "artist"
        WScript.Echo track.Artist
        WScript.Echo "size"
        WScript.Echo track.Size
        WScript.Echo "rating"
        WScript.Echo track.Rating
    next
    WScript.Echo "*Done*"
    WScript.quit
end if

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' Reads ratings from StdIn, so to import your ratings, do something like this:
''''  Playlists.vbs iTunes importRatings <ratings.txt
'' Doesn't do much error checking, so be sure you have a well-formatted input file.
''
set objStdIn = WScript.StdIn

if optionImportRatings then
    '' define special constant for search cmd
    ITPlaylistSearchFieldSongNames = 5
    
    set playlist = playlists.Item(1)
    
    '' Skip over any random text at the start of the file
    strLine = ""
    do while StrLine <> "*New*"
        strLine = objStdIn.ReadLine()
        if strLine = "*Done*" then
            WScript.quit
        end if
    loop
    
    '' read data from Stdin, parsing it (I use the term loosely), and
    '' then try to find a matching track.
    allDone = false
    do while not allDone
        '' Found some data in the file
        trackName   = ""
        trackAlbum  = ""
        trackArtist = ""
        trackSize   = ""
        entryDone = false
        do while not entryDone
            someKindOfData = objStdIn.ReadLine()
            if someKindOfData = "name" then
                trackName = objStdIn.ReadLine()
            elseif someKindOfData = "album" then
                trackAlbum = objStdIn.ReadLine()
            elseif someKindOfData = "artist" then
                trackArtist = objStdIn.ReadLine()
            elseif someKindOfData = "size" then
                trackSize = objStdIn.ReadLine()
            elseif someKindOfData = "rating" then
                trackRating = objStdIn.ReadLine()
            elseif someKindOfData = "*New*" then
                entryDone = true
            elseif someKindOfData = "*Done*" then
                allDone   = true
                entryDone = true
            else
                WScript.Echo "Unknown field: " & someKindOfData
            end if

        loop

        '' Search song names, then try to refine the match
        '' This "Search" function is like the one in iTunes - it's more of a filter,
        '' so it can (and does) return a list of tracks with the same substring
        '' in the name.

        set possibleTracks = playlist.Search(trackName, ITPlaylistSearchFieldSongNames)
    
        if IsObject(possibleTracks) and not IsNull(possibleTracks) then
            trackCount = possibleTracks.Count
    
            for i = 1 to trackCount
                set track = possibleTracks.Item(i)
                '' Only check out tracks where the fields match exactly ...
                '' [I had a bug where I used CInt on the track.Size - the range of a CInt is very small (16-bits I assume),
                ''  and if the conversion overflows you just get a blank result, and both compares will succeed!]
                modifyRating = false
                if optionCheckName and track.Name = trackName then
                    if optionVerbose then
                        WScript.Echo "Matched " & trackName
                    end if
                    modifyRating = True
                end if
                if optionCheckSize and CLng(track.Size) = CLng(trackSize) then
                    if optionVerbose then
                        WScript.Echo "Matched " & trackSize
                    end if
                    modifyRating = modifyRating and true
                end if
                if optionCheckAlbum and track.Album = trackAlbum then
                    if optionVerbose then
                        WScript.Echo "Matched " & trackAlbum
                    end if
                    modifyRating = modifyRating and true
                end if
                if optionCheckSize and track.Artist = trackArtist then
                    if optionVerbose then
                        WScript.Echo "Matched " & trackArtist
                    end if
                    modifyRating = modifyRating and true
                end if
                if modifyRating then
                    '' If the entry in the library has a non-zero rating and the text file 
                    '' rating is zero, keep the library rating, because a zero rating 
                    '' means "not rated", so why clobber the rating already in iTunes?
                    if CInt(trackRating) <> 0 and CInt(trackRating) <> CInt(track.Rating) then
                        if not optionTestOnly then
                            track.Rating = CInt(trackRating)
                            WScript.Echo "Really setting " & trackName & " from " & track.Rating & " to " & trackRating
                        else
                            WScript.Echo "Pretend setting " & trackName & " from " & track.Rating & " to " & trackRating
                        end if
                    else
                        if optionVerbose then
                            WScript.Echo "Skipping " & trackName & ":  Same rating."
                        end if
                    end if
                else
                    if optionVerbose then
                        WScript.Echo "Skipping " & trackName & " - not matched."
                        WScript.Echo "Checked "         & optionCheckName    & " Input " & trackName     & " - vs - lib " & track.Name
                        WScript.Echo "Checked "         & optionCheckArtist  & " Input " & trackArtist   & " - vs - lib " & track.Artist
                        WScript.Echo "Checked "         & optionCheckAlbum   & " Input " & trackAlbum    & " - vs - lib " & track.Album
                        WScript.Echo "Checked "         & optionCheckSize    & " Input " & trackSize     & " - vs - lib " & track.Size
                    end if
                end if
            next
        end if
    loop
    WScript.quit
end if


''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' The remaining options, which deal with unreferenced tracks,
''  are interleaved in the code below, since they all need a 
'' similar loop.

WScript.Echo playlistsCount & " playlists"

for i = 1 to playlistsCount  ' the first playlist is everything on the iPod (or in iTunes)
    
    okay_continue = true
    set playlist = playlists.Item(i)

    if playlist.Kind = ITPlaylistKindLibrary then
        WScript.Echo "Main = " & playlist.Name
    elseif playlist.Kind <> ITPlaylistKindUser then
        WScript.Echo "Skipping non-user playlist " & playlist.Name
        okay_continue = false
    elseif playlist.Smart and optionIgnoreSmartPlaylists then
        WScript.Echo "Skipping smart playlist " & playlist.Name
        okay_continue = false
    end if

    '' If there is an existing "Unreferenced" playlist, remember it, in case it's needed later
    if playlist.Name = "Unreferenced" then
        set unreferencedPlaylist = playlist
    end if
    
    if okay_continue then 
        
        set tracks  = playlist.Tracks
        numTracks   = tracks.Count
        
        for j = 1 to numTracks
            
            set track   = tracks.Item(j)
            album       = track.Album
            artist      = track.Artist
            rating      = track.Rating
            trackName   = track.Name
            
            if optionVerbose then
                WScript.Echo "objectIdx = " & objectIdx & " Playlist: '" & playlist.Name & "'   Trackname: '" & trackName & "'" & " Rating: " & track.Rating
            end if
            
            if not trackCollection.Exists(trackName) then
                set trackItem = CreateObject("Scripting.Dictionary")
                trackItem.Add "Count", 0
                trackCollection.Add trackName, trackItem
            else
                set trackItem = trackCollection.Item(trackName)
            end if
            
            '' Bump count of number of times this track has appeared in a playlist
            count = trackItem.Item("Count") + 1
            trackItem.Item("Count") = count

            '' Save this particular track under the 'count' entry
            trackItem.Add count, objectIdx

            '' Save track Object in trackObjects array and bump objectIdx
            set trackObjects(objectIdx) = track
            objectIdx = objectIdx + 1
            if objectIdx >= UBound(trackObjects) then
                WScript.echo "trackObjects array overflow (" & objectIdx & ")"
                WScript.quit
            end if
            
        Next
    end if 'okay_continue
next

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'' If we get here, then the default behavior is to print out tracks that
'' do not show up on any playlists (not counting the main library), and
'' optionally delete them.
''
'' This was so I could not find unreferenced Podcasts and such on my iPod
'' because my 3G iPod would lose track of stuff (never to be found again
'' without this script).  Things really broke when they introduced the separate
'' Podcast playlist, which is basically broken on a 3G iPod.
''
WScript.Echo ""
WScript.Echo "Tracks that are not on any playlist:"
WScript.Echo ""

for each trackNameKey in trackCollection
    ''WScript.Echo "trackNameKey = " & trackNameKey
    set trackItem = trackCollection.Item(trackNameKey)
    count = trackItem.Item("Count")
    if count = 1 then
        
        objectIdx   = trackItem.Item(1)
        set track   = trackObjects(objectIdx)

        name            = track.Name
        album           = track.Album
        artist          = track.Artist
        rating          = track.Rating
        
        WScript.Echo "Playlist: 'None'   Trackname: '" & name & "'; Artist: '" & artist & "'; Album: '" & album & "'; Rating: " & rating
        
        if optionMoveToUnreferencedList then
            if not IsObject(unreferencedPlaylist) then
                if optionSourceType = ITSourceKindLibrary then
                    set unreferencedPlaylist = iTunesApp.CreatePlayList("Unreferenced")
                else
                    WScript.Echo "Can only automatically create 'Unreferenced' playlist in iTunes."
                    WScript.Echo "But you can create it manually on your iPod if you want."
                    Wscript.Echo "Aborting."
                    WScript.Quit
                end if
            end if
            WScript.Echo "Moving to 'Unreferenced' playlist; objectIdx = " & objectIdx & " track = " & track.Name
            unreferencedPlaylist.AddTrack(track)
        elseif optionDeleteUnreferenced then
            WScript.Echo "Really deleting objectIdx = " & objectIdx & " track = " & track.Name
            if not optionTestOnly then
                track.Delete
            end if
        end if

    end if '' count = 1
next '' trackNameKey

